@@ -1,7 +1,7 @@
 #!/usr/bin/env python3
 # SPDX-License-Identifier: Apache-2.0
 # Copyright 2018-2021 Michael R. Crusoe
-"""CWL Expression refactoring tool for CWL v1.0 ."""
+"""CWL Expression refactoring tool for CWL v1.2 ."""
 import copy
 import hashlib
 from collections.abc import Mapping
@@ -22,7 +22,8 @@ from ruamel import yaml
 from schema_salad.sourceline import SourceLine
 from schema_salad.utils import json_dumps
 
-import cwl_utils.parser.cwl_v1_0 as cwl
+import cwl_utils.parser.cwl_v1_2 as cwl
+import cwl_utils.parser.cwl_v1_2_utils as utils
 from cwl_utils.errors import JavascriptException, WorkflowException
 from cwl_utils.expression import do_eval, interpolate
 from cwl_utils.types import CWLObjectType, CWLOutputType
@@ -182,7 +183,7 @@ var runtime=$(runtime);"""
         contents += "\n" + "\n".join(expressionLib)
     contents += (
         """
-var ret = function(){"""
+    var ret = function(){"""
         + escape_expression_field(etool.expression.strip()[2:-1])
         + """}();
 process.stdout.write(JSON.stringify(ret));"""
@@ -235,7 +236,7 @@ def traverse(
                     )
                 )
                 wf_inputs.append(
-                    cwl.InputParameter(
+                    cwl.WorkflowInputParameter(
                         id=inp_id,
                         label=inp.label,
                         secondaryFiles=inp.secondaryFiles,
@@ -323,13 +324,13 @@ def load_step(
 
 def generate_etool_from_expr(
     expr: str,
-    target: Union[cwl.CommandInputParameter, cwl.InputParameter],
+    target: Union[cwl.CommandInputParameter, cwl.WorkflowInputParameter],
     no_inputs: bool = False,
     self_type: Optional[
         Union[
-            cwl.InputParameter,
+            cwl.WorkflowInputParameter,
             cwl.CommandInputParameter,
-            List[Union[cwl.InputParameter, cwl.CommandInputParameter]],
+            List[Union[cwl.WorkflowInputParameter, cwl.CommandInputParameter]],
         ]
     ] = None,  # if the "self" input should be a different type than the "result" output
     extra_processes: Optional[
@@ -345,13 +346,11 @@ def generate_etool_from_expr(
             new_type: Union[
                 List[Union[cwl.ArraySchema, cwl.InputRecordSchema]],
                 Union[cwl.ArraySchema, cwl.InputRecordSchema],
-            ] = [clean_type_ids(t.type) for t in self_type if t.type]
-        elif self_type.type:
-            new_type = clean_type_ids(self_type.type)
+            ] = [clean_type_ids(t.type) for t in self_type]
         else:
-            raise WorkflowException(f"Don't know how to make type from '{self_type}'.")
+            new_type = clean_type_ids(self_type.type)
         inputs.append(
-            cwl.InputParameter(
+            cwl.WorkflowInputParameter(
                 id="self",
                 label=self_type.label if not isinstance(self_type, list) else None,
                 secondaryFiles=self_type.secondaryFiles
@@ -390,7 +389,7 @@ def generate_etool_from_expr(
         expression += "\n  var self=inputs.self;"
     expression += (
         """
-  return {"result": function(){"""
+      return {"result": function(){"""
         + expr[2:-2]
         + """}()};
  }"""
@@ -449,7 +448,7 @@ def replace_expr_with_etool(
     expr: str,
     name: str,
     workflow: cwl.Workflow,
-    target: Union[cwl.CommandInputParameter, cwl.InputParameter],
+    target: Union[cwl.CommandInputParameter, cwl.WorkflowInputParameter],
     source: Optional[Union[str, List[Any]]],
     replace_etool: bool = False,
     extra_process: Optional[
@@ -537,7 +536,7 @@ def empty_inputs(
             else:
                 try:
                     result[param_id] = example_input(
-                        type_for_source(process_or_step.run, param.source, parent)
+                        utils.type_for_source(process_or_step.run, param.source, parent)
                     )
                 except WorkflowException:
                     pass
@@ -582,71 +581,6 @@ def example_input(some_type: Any) -> Any:
     return None
 
 
-def type_for_source(
-    process: Union[cwl.CommandLineTool, cwl.Workflow, cwl.ExpressionTool],
-    sourcenames: Union[str, List[str]],
-    parent: Optional[cwl.Workflow] = None,
-) -> Union[List[Any], Any]:
-    """Determine the type for the given sourcenames."""
-    params = param_for_source_id(process, sourcenames, parent)
-    if not isinstance(params, list):
-        return params.type
-    new_type: List[Any] = []
-    for p in params:
-        if isinstance(p, str) and p not in new_type:
-            new_type.append(p)
-        elif hasattr(p, "type") and p.type not in new_type:
-            new_type.append(p.type)
-    return new_type
-
-
-def param_for_source_id(
-    process: Union[cwl.CommandLineTool, cwl.Workflow, cwl.ExpressionTool],
-    sourcenames: Union[str, List[str]],
-    parent: Optional[cwl.Workflow] = None,
-) -> Union[List[cwl.InputParameter], cwl.InputParameter]:
-    """Find the process input parameter that matches one of the given sourcenames."""
-    if isinstance(sourcenames, str):
-        sourcenames = [sourcenames]
-    params: List[cwl.InputParameter] = []
-    for sourcename in sourcenames:
-        if not isinstance(process, cwl.Workflow):
-            for param in process.inputs:
-                if param.id.split("#")[-1] == sourcename.split("#")[-1]:
-                    params.append(param)
-        targets = [process]
-        if parent:
-            targets.append(parent)
-        for target in targets:
-            if isinstance(target, cwl.Workflow):
-                for inp in target.inputs:
-                    if inp.id.split("#")[-1] == sourcename.split("#")[-1]:
-                        params.append(inp)
-                for step in target.steps:
-                    if sourcename.split("/")[0] == step.id.split("#")[-1] and step.out:
-                        for outp in step.out:
-                            outp_id = outp if isinstance(outp, str) else outp.id
-                            if outp_id.split("/")[-1] == sourcename.split("/", 1)[1]:
-                                if step.run and step.run.outputs:
-                                    for output in step.run.outputs:
-                                        if (
-                                            output.id.split("#")[-1]
-                                            == sourcename.split("/", 1)[1]
-                                        ):
-                                            params.append(output)
-    if len(params) == 1:
-        return params[0]
-    elif len(params) > 1:
-        return params
-    raise WorkflowException(
-        "param {} not found in {}\n or\n {}.".format(
-            sourcename,
-            yaml.main.round_trip_dump(cwl.save(process)),
-            yaml.main.round_trip_dump(cwl.save(parent)),
-        )
-    )
-
-
 EMPTY_FILE: CWLOutputType = {
     "class": "File",
     "basename": "em.pty",
@@ -667,6 +601,50 @@ TOPLEVEL_FORMAT_EXPR_ERROR = (
     "into separate ExpressionTool/CommandLineTool steps."
 )
 
+PICKVALUE_FIRST_NON_NULL_EXPR = """${
+for (let i = 0; i < self.length; i++) {
+  if (self[i] !== null){
+    return self[i];
+  }
+}
+throw 'pickValue=first_non_null, but no non-null value found: ' + self;
+}
+"""
+
+PICKVALUE_THE_ONLY_NON_NULL_EXPR = """${
+var found = null;
+self.forEach(function(item) {
+  if (item !== null) {
+    if (found !== null) {
+      throw 'pickValue=the_only_non_null, but multiple non-null values found: ' + self;
+    } else {
+      found = item;
+    }
+  }
+})
+if (found !== null) {
+  return found;
+}
+throw 'pickValue=the_only_non_null, but no non-null value found: ' + self;
+}
+"""
+
+PICKVALUE_ALL_NON_NULL_EXPR = """${
+var results = [];
+self.forEach(function(item) {
+  if (item !== null){
+    results.push(item);
+  }
+})
+return results;
+}
+"""
+
+PICKVALUE_ERROR = (
+    "pickValue '{}' is invalid. Should be one of 'first_non_null', "
+    "'the_only_non_null', 'all_non_null'"
+)
+
 
 def process_workflow_inputs_and_outputs(
     workflow: cwl.Workflow, replace_etool: bool
@@ -683,7 +661,9 @@ def process_workflow_inputs_and_outputs(
                     raise_type=WorkflowException,
                 ).makeError(TOPLEVEL_FORMAT_EXPR_ERROR.format(param.id.split("#")[-1]))
             if param.secondaryFiles:
-                if get_expression(param.secondaryFiles, inputs, EMPTY_FILE):
+                if hasattr(param.secondaryFiles, "pattern") and get_expression(
+                    param.secondaryFiles.pattern, inputs, EMPTY_FILE
+                ):
                     raise SourceLine(
                         param.loadingOptions.original_doc,
                         "secondaryFiles",
@@ -691,7 +671,7 @@ def process_workflow_inputs_and_outputs(
                     ).makeError(TOPLEVEL_SF_EXPR_ERROR.format(param.id.split("#")[-1]))
                 elif isinstance(param.secondaryFiles, MutableSequence):
                     for index2, entry in enumerate(param.secondaryFiles):
-                        if get_expression(entry, inputs, EMPTY_FILE):
+                        if get_expression(entry.pattern, inputs, EMPTY_FILE):
                             raise SourceLine(
                                 param.loadingOptions.original_doc,
                                 index2,
@@ -700,6 +680,58 @@ def process_workflow_inputs_and_outputs(
                                 f"Entry {index},"
                                 + TOPLEVEL_SF_EXPR_ERROR.format(param.id.split("#")[-1])
                             )
+    generated_pickValue_steps = []
+    for index3, param2 in enumerate(workflow.outputs):
+        with SourceLine(workflow.outputs, index3, WorkflowException):
+            if param2.pickValue:
+                if param2.pickValue == "first_non_null":
+                    expression = PICKVALUE_FIRST_NON_NULL_EXPR
+                elif param2.pickValue == "the_only_non_null":
+                    expression = PICKVALUE_THE_ONLY_NON_NULL_EXPR
+                elif param2.pickValue == "all_non_null":
+                    expression = PICKVALUE_ALL_NON_NULL_EXPR
+                else:
+                    raise WorkflowException(PICKVALUE_ERROR.format(param2.pickValue))
+                modified = True
+
+                etool_id = "_pickValue_workflow_step_{}".format(
+                    param2.id.split("#")[-1]
+                )
+                target_type = copy.deepcopy(param2.type)
+                if isinstance(target_type, cwl.OutputArraySchema):
+                    target_type.name = None
+                target = cwl.WorkflowInputParameter(id=None, type=target_type)
+                if not isinstance(param2.outputSource, list):
+                    sources: Union[List[str], str] = "/".join(
+                        param2.outputSource.split("#")[-1].split("/")[1:]
+                    )
+                else:
+                    sources = [
+                        "/".join(s.split("#")[-1].split("/")[1:])
+                        for s in param2.outputSource
+                    ]
+                source_type_items = utils.type_for_source(workflow, sources)
+                if "null" not in source_type_items:
+                    if isinstance(source_type_items, list):
+                        source_type_items.append("null")
+                    else:
+                        source_type_items = ["null", source_type_items]
+                source_type = cwl.CommandInputParameter(
+                    type=cwl.ArraySchema(type="array", items=source_type_items)
+                )
+                replace_expr_with_etool(
+                    expression,
+                    etool_id,
+                    workflow,
+                    target,
+                    sources,
+                    replace_etool,
+                    None,
+                    source_type,
+                )
+                param2.outputSource = f"{etool_id}/result"
+                param2.pickValue = None
+                generated_pickValue_steps.append(etool_id)
     return modified
 
 
@@ -742,7 +774,7 @@ def process_workflow_reqs_and_hints(
                             expression = get_expression(envDef.envValue, inputs, None)
                             if expression:
                                 modified = True
-                                target = cwl.InputParameter(
+                                target = cwl.WorkflowInputParameter(
                                     id=None,
                                     type="string",
                                 )
@@ -773,7 +805,7 @@ def process_workflow_reqs_and_hints(
                         expression = get_expression(this_attr, inputs, None)
                         if expression:
                             modified = True
-                            target = cwl.InputParameter(id=None, type="long")
+                            target = cwl.WorkflowInputParameter(id=None, type="long")
                             etool_id = (
                                 "_expression_workflow_ResourceRequirement_{}".format(
                                     attr
@@ -800,7 +832,7 @@ def process_workflow_reqs_and_hints(
                         expression = get_expression(req.listing, inputs, None)
                         if expression:
                             modified = True
-                            target = cwl.InputParameter(
+                            target = cwl.WorkflowInputParameter(
                                 id=None,
                                 type=cwl.InputArraySchema(
                                     ["File", "Directory"], "array", None, None
@@ -826,7 +858,7 @@ def process_workflow_reqs_and_hints(
                             expression = get_expression(entry, inputs, None)
                             if expression:
                                 modified = True
-                                target = cwl.InputParameter(
+                                target = cwl.WorkflowInputParameter(
                                     id=None,
                                     type=cwl.InputArraySchema(
                                         ["File", "Directory"], "array", None, None
@@ -869,7 +901,7 @@ def process_workflow_reqs_and_hints(
                                                 or expr_result["class"] == "Directory"
                                             )
                                         ):
-                                            target = cwl.InputParameter(
+                                            target = cwl.WorkflowInputParameter(
                                                 id=None,
                                                 type=expr_result["class"],
                                             )
@@ -890,7 +922,7 @@ def process_workflow_reqs_and_hints(
                                                 (etool_id, index)
                                             )
                                         elif isinstance(expr_result, str):
-                                            target = cwl.InputParameter(
+                                            target = cwl.WorkflowInputParameter(
                                                 id=None,
                                                 type=["File"],
                                             )
@@ -932,7 +964,7 @@ def process_workflow_reqs_and_hints(
                                     )
                                     if expression:
                                         modified = True
-                                        target = cwl.InputParameter(
+                                        target = cwl.WorkflowInputParameter(
                                             id=None,
                                             type="string",
                                         )
@@ -1063,7 +1095,7 @@ def process_level_reqs(
                         expression = get_expression(envDef.envValue, inputs, None)
                         if expression:
                             modified = True
-                            target = cwl.InputParameter(id=None, type="string")
+                            target = cwl.WorkflowInputParameter(id=None, type="string")
                             etool_id = "_expression_{}_EnvVarRequirement_{}".format(
                                 step_name, env_index
                             )
@@ -1087,7 +1119,7 @@ def process_level_reqs(
                     expression = get_expression(this_attr, inputs, None)
                     if expression:
                         modified = True
-                        target = cwl.InputParameter(id=None, type="long")
+                        target = cwl.WorkflowInputParameter(id=None, type="long")
                         etool_id = "_expression_{}_ResourceRequirement_{}".format(
                             step_name, attr
                         )
@@ -1114,7 +1146,7 @@ def process_level_reqs(
                         target_type = cwl.InputArraySchema(
                             ["File", "Directory"], "array", None, None
                         )
-                        target = cwl.InputParameter(id=None, type=target_type)
+                        target = cwl.WorkflowInputParameter(id=None, type=target_type)
                         etool_id = "_expression_{}_InitialWorkDirRequirement".format(
                             step_name
                         )
@@ -1150,7 +1182,7 @@ def process_level_reqs(
                             target_type = cwl.InputArraySchema(
                                 ["File", "Directory"], "array", None, None
                             )
-                            target = cwl.InputParameter(
+                            target = cwl.WorkflowInputParameter(
                                 id=None,
                                 type=target_type,
                             )
@@ -1207,7 +1239,7 @@ return result; }"""
                                     else:
                                         new_expression = expression
                                     d_target_type = ["File", "Directory"]
-                                    target = cwl.InputParameter(
+                                    target = cwl.WorkflowInputParameter(
                                         id=None,
                                         type=d_target_type,
                                     )
@@ -1237,7 +1269,7 @@ return result; }"""
                                 )
                                 if expression:
                                     modified = True
-                                    target = cwl.InputParameter(
+                                    target = cwl.WorkflowInputParameter(
                                         id=None,
                                         type="string",
                                     )
@@ -1315,7 +1347,7 @@ def traverse_CommandLineTool(
                     inp_id = f"_arguments_{index}"
                     etool_id = f"_expression_{step_id}{inp_id}"
                     target_type = "Any"
-                    target = cwl.InputParameter(id=None, type=target_type)
+                    target = cwl.WorkflowInputParameter(id=None, type=target_type)
                     replace_step_clt_expr_with_etool(
                         expression, etool_id, parent, target, step, replace_etool
                     )
@@ -1341,7 +1373,7 @@ def traverse_CommandLineTool(
                     inp_id = f"_arguments_{index}"
                     etool_id = f"_expression_{step_id}{inp_id}"
                     target_type = "Any"
-                    target = cwl.InputParameter(id=None, type=target_type)
+                    target = cwl.WorkflowInputParameter(id=None, type=target_type)
                     replace_step_clt_expr_with_etool(
                         expression, etool_id, parent, target, step, replace_etool
                     )
@@ -1367,7 +1399,7 @@ def traverse_CommandLineTool(
                 inp_id = f"_{streamtype}"
                 etool_id = f"_expression_{step_id}{inp_id}"
                 target_type = "string"
-                target = cwl.InputParameter(id=None, type=target_type)
+                target = cwl.WorkflowInputParameter(id=None, type=target_type)
                 replace_step_clt_expr_with_etool(
                     expression, etool_id, parent, target, step, replace_etool
                 )
@@ -1407,7 +1439,7 @@ def traverse_CommandLineTool(
                     inp_id = "_{}_glob".format(outp.id.split("#")[-1])
                     etool_id = f"_expression_{step_id}{inp_id}"
                     glob_target_type = ["string", cwl.ArraySchema("string", "array")]
-                    target = cwl.InputParameter(id=None, type=glob_target_type)
+                    target = cwl.WorkflowInputParameter(id=None, type=glob_target_type)
                     replace_step_clt_expr_with_etool(
                         expression, etool_id, parent, target, step, replace_etool
                     )
@@ -1443,7 +1475,7 @@ def traverse_CommandLineTool(
                     sub_wf_outputs = cltool_step_outputs_to_workflow_outputs(
                         step, etool_id, outp_id
                     )
-                    self_type = cwl.InputParameter(
+                    self_type = cwl.WorkflowInputParameter(
                         id=None,
                         type=cwl.InputArraySchema("File", "array", None, None),
                     )
@@ -1620,7 +1652,7 @@ def replace_step_clt_expr_with_etool(
     expr: str,
     name: str,
     workflow: cwl.Workflow,
-    target: cwl.InputParameter,
+    target: cwl.WorkflowInputParameter,
     step: cwl.WorkflowStep,
     replace_etool: bool,
     self_name: Optional[str] = None,
@@ -1655,7 +1687,7 @@ def replace_clt_hintreq_expr_with_etool(
     expr: str,
     name: str,
     workflow: cwl.Workflow,
-    target: cwl.InputParameter,
+    target: cwl.WorkflowInputParameter,
     step: cwl.WorkflowStep,
     replace_etool: bool,
     self_name: Optional[str] = None,
@@ -1690,7 +1722,7 @@ def replace_clt_hintreq_expr_with_etool(
 
 def cltool_inputs_to_etool_inputs(
     tool: cwl.CommandLineTool,
-) -> List[cwl.InputParameter]:
+) -> List[cwl.WorkflowInputParameter]:
     """Copy CommandLineTool input objects into the equivalent ExpressionTool input objects."""
     inputs = yaml.comments.CommentedSeq()
     if tool.inputs:
@@ -1698,7 +1730,7 @@ def cltool_inputs_to_etool_inputs(
             clt_inp_id = clt_inp.id.split("#")[-1].split("/")[-1]
             if not clt_inp_id.startswith("_"):
                 inputs.append(
-                    cwl.InputParameter(
+                    cwl.WorkflowInputParameter(
                         id=clt_inp_id,
                         label=clt_inp.label,
                         secondaryFiles=clt_inp.secondaryFiles,
@@ -1754,8 +1786,8 @@ def cltool_step_outputs_to_workflow_outputs(
 
 def generate_etool_from_expr2(
     expr: str,
-    target: cwl.InputParameter,
-    inputs: Sequence[Union[cwl.InputParameter, cwl.CommandInputParameter]],
+    target: Union[cwl.CommandInputParameter, cwl.WorkflowInputParameter],
+    inputs: Sequence[Union[cwl.WorkflowInputParameter, cwl.CommandInputParameter]],
     self_name: Optional[str] = None,
     process: Optional[Union[cwl.CommandLineTool, cwl.ExpressionTool]] = None,
     extra_processes: Optional[
@@ -1780,7 +1812,7 @@ def generate_etool_from_expr2(
         expression += f"\n  var self=inputs.{self_name};"
     expression += (
         """
-  return {"result": function(){"""
+      return {"result": function(){"""
         + expr[2:-2]
         + """}()};
  }"""
@@ -1841,11 +1873,13 @@ def traverse_step(
                         if not step.scatter:
                             self.append(
                                 example_input(
-                                    type_for_source(parent, source.split("#")[-1])
+                                    utils.type_for_source(parent, source.split("#")[-1])
                                 )
                             )
                         else:
-                            scattered_source_type = type_for_source(parent, source)
+                            scattered_source_type = utils.type_for_source(
+                                parent, source
+                            )
                             if isinstance(scattered_source_type, list):
                                 for stype in scattered_source_type:
                                     self.append(example_input(stype.type))
@@ -1854,10 +1888,12 @@ def traverse_step(
                 else:
                     if not step.scatter:
                         self = example_input(
-                            type_for_source(parent, inp.source.split("#")[-1])
+                            utils.type_for_source(parent, inp.source.split("#")[-1])
                         )
                     else:
-                        scattered_source_type2 = type_for_source(parent, inp.source)
+                        scattered_source_type2 = utils.type_for_source(
+                            parent, inp.source
+                        )
                         if isinstance(scattered_source_type2, list):
                             self = example_input(scattered_source_type2[0].type)
                         else:
@@ -1871,16 +1907,18 @@ def traverse_step(
                     raise WorkflowException("target not found")
                 input_source_id = None
                 source_type: Optional[
-                    Union[List[cwl.InputParameter], cwl.InputParameter]
+                    Union[List[cwl.WorkflowInputParameter], cwl.WorkflowInputParameter]
                 ] = None
                 if inp.source:
                     if isinstance(inp.source, MutableSequence):
                         input_source_id = []
-                        source_types: List[cwl.InputParameter] = []
+                        source_types: List[cwl.WorkflowInputParameter] = []
                         for source in inp.source:
                             source_id = source.split("#")[-1]
                             input_source_id.append(source_id)
-                            temp_type = type_for_source(step.run, source_id, parent)
+                            temp_type = utils.type_for_source(
+                                step.run, source_id, parent
+                            )
                             if isinstance(temp_type, list):
                                 for ttype in temp_type:
                                     if ttype not in source_types:
@@ -1888,13 +1926,13 @@ def traverse_step(
                             else:
                                 if temp_type not in source_types:
                                     source_types.append(temp_type)
-                        source_type = cwl.InputParameter(
+                        source_type = cwl.WorkflowInputParameter(
                             id=None,
                             type=cwl.ArraySchema(source_types, "array"),
                         )
                     else:
                         input_source_id = inp.source.split("#")[-1]
-                        source_type = param_for_source_id(
+                        source_type = utils.param_for_source_id(
                             step.run, input_source_id, parent
                         )
                 # target.id = target.id.split('#')[-1]
@@ -1930,6 +1968,14 @@ def traverse_step(
                 )
                 inp.valueFrom = None
                 inp.source = f"{etool_id}/result"
+    if step.when:
+        expression = get_expression(string=step.when, inputs=inputs, self=None)
+        if expression:
+            modified = True
+            replace_step_when_expr_with_etool(
+                expression, parent, step, original_step_ins, replace_etool
+            )
+
     # TODO: skip or special process for sub workflows?
     process_modified = process_level_reqs(
         original_process,
@@ -1955,31 +2001,25 @@ def traverse_step(
     return modified
 
 
-def workflow_step_to_InputParameters(
+def workflow_step_to_WorkflowInputParameters(
     step_ins: List[cwl.WorkflowStepInput], parent: cwl.Workflow, except_in_id: str
-) -> List[cwl.InputParameter]:
-    """Create InputParameters to match the given WorkflowStep inputs."""
+) -> List[cwl.WorkflowInputParameter]:
+    """Create WorkflowInputParameters to match the given WorkflowStep inputs."""
     params = []
     for inp in step_ins:
         if not inp.id:
             continue
         inp_id = inp.id.split("#")[-1].split("/")[-1]
         if inp.source and inp_id != except_in_id:
-            param = copy.deepcopy(param_for_source_id(parent, sourcenames=inp.source))
+            param = copy.deepcopy(
+                utils.param_for_source_id(parent, sourcenames=inp.source)
+            )
             if isinstance(param, list):
                 for p in param:
-                    if not p.type:
-                        raise WorkflowException(
-                            f"Don't know how to get type id for '{p}'."
-                        )
                     p.id = inp_id
                     p.type = clean_type_ids(p.type)
                     params.append(p)
             else:
-                if not param.type:
-                    raise WorkflowException(
-                        f"Don't know how to get type id for '{param}'."
-                    )
                 param.id = inp_id
                 param.type = clean_type_ids(param.type)
                 params.append(param)
@@ -1990,24 +2030,26 @@ def replace_step_valueFrom_expr_with_etool(
     expr: str,
     name: str,
     workflow: cwl.Workflow,
-    target: Union[cwl.CommandInputParameter, cwl.InputParameter],
+    target: Union[cwl.CommandInputParameter, cwl.WorkflowInputParameter],
     step: cwl.WorkflowStep,
     step_inp: cwl.WorkflowStepInput,
     original_process: Union[cwl.CommandLineTool, cwl.ExpressionTool],
     original_step_ins: List[cwl.WorkflowStepInput],
     source: Union[str, List[str]],
     replace_etool: bool,
-    source_type: Optional[Union[cwl.InputParameter, List[cwl.InputParameter]]] = None,
+    source_type: Optional[
+        Union[cwl.WorkflowInputParameter, List[cwl.WorkflowInputParameter]]
+    ] = None,
 ) -> None:
     """Replace a WorkflowStep level 'valueFrom' expression with a sibling ExpressionTool step."""
     if not step_inp.id:
         raise WorkflowException(f"Missing id in {step_inp}.")
     step_inp_id = step_inp.id.split("/")[-1]
-    etool_inputs = workflow_step_to_InputParameters(
+    etool_inputs = workflow_step_to_WorkflowInputParameters(
         original_step_ins, workflow, step_inp_id
     )
     if source:
-        source_param = cwl.InputParameter(id="self", type="Any")
+        source_param = cwl.WorkflowInputParameter(id="self", type="Any")
         # TODO: would be nicer to derive a proper type; but in the face of linkMerge, this is easier for now
         etool_inputs.append(source_param)
     temp_etool = generate_etool_from_expr2(
@@ -2031,6 +2073,8 @@ def replace_step_valueFrom_expr_with_etool(
     if source:
         wf_step_inputs.append(cwl.WorkflowStepInput(id="self", source=step_inp.source))
     for wf_step_input in wf_step_inputs:
+        if not wf_step_input.id:
+            continue
         wf_step_input.id = wf_step_input.id.split("/")[-1]
         if wf_step_input.valueFrom:
             wf_step_input.valueFrom = None
@@ -2068,6 +2112,70 @@ def replace_step_valueFrom_expr_with_etool(
     )
 
 
+def replace_step_when_expr_with_etool(
+    expr: str,
+    workflow: cwl.Workflow,
+    step: cwl.WorkflowStep,
+    original_step_ins: List[cwl.WorkflowStepInput],
+    replace_etool: bool,
+) -> None:
+    """Replace a WorkflowStep level 'when' expression with a sibling ExpressionTool step."""
+    if not step.id:
+        raise WorkflowException(f"Missing id from {step}.")
+    etool_id = "_when_expression_{}".format(step.id.split("#")[-1])
+    etool_inputs = workflow_step_to_WorkflowInputParameters(
+        original_step_ins, workflow, ""
+    )
+    temp_etool = generate_etool_from_expr2(
+        expr,
+        cwl.WorkflowInputParameter(id=None, type="boolean"),
+        etool_inputs,
+        None,
+        None,
+        [workflow, step],
+    )
+    if replace_etool:
+        processes = [
+            workflow,
+            step,
+        ]  # type: List[Union[cwl.Workflow, cwl.CommandLineTool, cwl.ExpressionTool, cwl.WorkflowStep]]
+        cltool = etool_to_cltool(temp_etool, find_expressionLib(processes))
+        etool = cltool  # type: Union[cwl.ExpressionTool, cwl.CommandLineTool]
+    else:
+        etool = temp_etool
+    wf_step_inputs = copy.deepcopy(original_step_ins)
+    for wf_step_input in wf_step_inputs:
+        if not wf_step_input.id:
+            continue
+        wf_step_input.id = wf_step_input.id.split("/")[-1]
+        if wf_step_input.source:
+            if isinstance(wf_step_input.source, MutableSequence):
+                for index, inp_source in enumerate(wf_step_input.source):
+                    wf_step_input.source[index] = inp_source.split("#")[-1]
+            else:
+                wf_step_input.source = wf_step_input.source.split("#")[-1]
+    wf_step_inputs[:] = [x for x in wf_step_inputs if x.id and not x.id.startswith("_")]
+    scatter = copy.deepcopy(step.scatter)
+    if isinstance(scatter, str):
+        scatter = [scatter]
+    if isinstance(scatter, MutableSequence):
+        for index, entry in enumerate(scatter):
+            scatter[index] = entry.split("/")[-1]
+    scatter = step.scatter
+    workflow.steps.append(
+        cwl.WorkflowStep(
+            id=etool_id,
+            in_=wf_step_inputs,
+            out=[cwl.WorkflowStepOutput("result")],
+            run=etool,
+            scatter=scatter,
+            scatterMethod=step.scatterMethod,
+        )
+    )
+    step.when = "$(inputs._when)"
+    step.in_.append(cwl.WorkflowStepInput(id="_when", source=f"{etool_id}/result"))
+
+
 def traverse_workflow(
     workflow: cwl.Workflow,
     replace_etool: bool,
